import json
from pathlib import Path
from typing import Any


def read_json(file_path: Path) -> Any:
    """Read content in a json file"""
    with open(file_path, "r") as file:
        return json.load(file)


def read_json_stream(stream: str) -> Any:
    """Read json formatted stream"""
    return json.loads(stream)


def write_json(content: Any, file_path: Path) -> None:
    """Write content to a json file"""
    with open(file_path, "w") as file:
        json.dump(content, file, indent=4)


def write_json_stream(content: Any) -> str:
    """Write content to a json stream"""
    return json.dumps(content)



from pydantic.dataclasses import dataclass


@dataclass
class RNADNACollection:
    """Contains the id for an RNA sample, the name of its connected DNA sample,
    and a list of connected, uploaded DNA cases."""

    rna_sample_id: str
    dna_sample_name: str
    dna_case_ids: list[str]



from pathlib import Path
from typing import Iterator

from cg.apps.lims import LimsAPI
from cg.constants.constants import GenomeVersion
from cg.constants.scout import ScoutExportFileName
from cg.exc import CgError
from cg.io.yaml import read_yaml, write_yaml_nextflow_style
from cg.models.tomte.tomte import TomteParameters
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.abstract import (
    ParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.utils import (
    replace_values_in_params_file,
)
from cg.store.store import Store


class TomteParamsFileCreator(ParamsFileCreator):
    def __init__(self, params: str, lims_api: LimsAPI, status_db: Store):
        super().__init__(params)
        self.lims_api = lims_api
        self.status_db = status_db

    def create(self, case_id: str, file_path: Path, sample_sheet_path: Path) -> None:
        content: dict = self._get_content(
            case_id=case_id,
            case_run_directory=file_path.parent,
            sample_sheet_path=sample_sheet_path,
        )
        write_yaml_nextflow_style(content=content, file_path=file_path)

    def _get_content(self, case_id: str, case_run_directory: Path, sample_sheet_path: Path):
        case_parameters = TomteParameters(
            input=sample_sheet_path,
            outdir=case_run_directory,
            gene_panel_clinical_filter=Path(case_run_directory, ScoutExportFileName.PANELS),
            tissue=self._get_case_source_type(case_id),  # type:ignore
            genome=GenomeVersion.HG38,
        ).model_dump()

        workflow_params = self._get_workflow_params()
        workflow_parameters: dict = workflow_params | case_parameters
        return replace_values_in_params_file(workflow_parameters=workflow_parameters)

    def _get_workflow_params(self) -> dict:
        return read_yaml(self.params)

    def _get_case_source_type(self, case_id: str) -> str | None:
        """
        Return the sample source type of a case.

        Raises:
            CgError: If different sources are set for the samples linked to a case.
        """
        sample_ids: Iterator[str] = self.status_db.get_sample_ids_by_case_id(case_id=case_id)
        source_types: set[str | None] = {
            self.lims_api.get_source(sample_id) for sample_id in sample_ids
        }
        if len(source_types) > 1:
            raise CgError(f"Different source types found for case: {case_id} ({source_types})")
        return source_types.pop()



from cg.constants.priority import SlurmQos
from cg.services.analysis_starter.configurator.abstract_model import CaseConfig


class NextflowCaseConfig(CaseConfig):
    case_priority: SlurmQos
    config_profiles: list[str]
    nextflow_config_file: str
    params_file: str
    pipeline_repository: str
    pre_run_script: str
    resume: bool = False
    revision: str
    session_id: str | None = None
    work_dir: str
    workflow_id: str | None = None

    def get_session_id(self) -> str | None:
        return self.session_id

    def get_workflow_id(self) -> str | None:
        return self.workflow_id



AUTHORS = [
    "Jan Albert",
    "Tobias Allander",
    "Sandra Broddesson",
    "Robert Dyrdak",
    "Martin Ekman",
    "Lynda Eneh",
    "Shambhu Ganeshappa Aralaguppe",
    "Natalija Gerasimcik",
    "Karina Hentrich",
    "Annika Tiveljung Lindell",
    "Valtteri Wirta",
    "Zhibing Yun",
]

HEADERS = [
    "submitter",
    "fn",
    "covv_virus_name",
    "covv_type",
    "covv_passage",
    "covv_collection_date",
    "covv_location",
    "covv_host",
    "covv_gender",
    "covv_patient_age",
    "covv_patient_status",
    "covv_seq_technology",
    "covv_orig_lab",
    "covv_orig_lab_addr",
    "covv_subm_sample_id",
    "covv_subm_lab",
    "covv_subm_lab_addr",
    "covv_authors",
    "covv_specimen",
    "covv_outbreak",
    "covv_add_host_info",
    "covv_add_location",
    "covv_provider_sample_id",
    "covv_last_vaccinated",
    "covv_treatment",
    "covv_assembly_method",
    "covv_coverage",
]



import logging
from datetime import datetime

from cg.constants.constants import CaseActions, DataDelivery, Workflow
from cg.constants.pedigree import Pedigree
from cg.services.orders.constants import ORDER_TYPE_WORKFLOW_MAP
from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.storing.service import StoreOrderService
from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.existing_case import ExistingCase
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.models.sample_aliases import SampleInCase
from cg.store.models import ApplicationVersion
from cg.store.models import Case as DbCase
from cg.store.models import CaseSample, Customer
from cg.store.models import Order as DbOrder
from cg.store.models import Sample as DbSample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StoreCaseOrderService(StoreOrderService):
    """
    Service for storing generic orders in StatusDB and Lims.
    This class is used to store orders for the following order types:
    - Balsamic
    - Balsamic UMI
    - MIP DNA
    - MIP RNA
    - Raredisease
    - RNAFusion
    - Tomte
    """

    def __init__(
        self,
        status_db: Store,
        lims_service: OrderLimsService,
    ):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: OrderWithCases) -> dict:
        """Submit a batch of samples for sequencing and analysis."""
        project_data = lims_map = None
        if new_samples := [sample for _, _, sample in order.enumerated_new_samples]:
            project_data, lims_map = self.lims.process_lims(
                samples=new_samples,
                customer=order.customer,
                ticket=order._generated_ticket_id,
                order_name=order.name,
                workflow=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
                delivery_type=order.delivery_type,
                skip_reception_control=order.skip_reception_control,
            )
        if lims_map:
            self._fill_in_sample_ids(samples=new_samples, lims_map=lims_map)

        new_cases: list[DbCase] = self.store_order_data_in_status_db(order)
        return {"project": project_data, "records": new_cases}

    def store_order_data_in_status_db(self, order: OrderWithCases) -> list[DbCase]:
        """Store cases, samples and their relationship in the Status database."""
        new_cases: list[DbCase] = []
        db_order = self._create_db_order(order)
        for case in order.cases:
            if case.is_new:
                db_case: DbCase = self._create_db_case(
                    case=case,
                    customer=db_order.customer,
                    ticket=str(order._generated_ticket_id),
                    workflow=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
                    delivery_type=order.delivery_type,
                )
                new_cases.append(db_case)
                self._update_case_panel(panels=getattr(case, "panels", []), case=db_case)
                case_samples: dict[str, DbSample] = self._create_db_sample_dict(
                    case=case, order=order, customer=db_order.customer
                )
                self._create_links(case=case, db_case=db_case, case_samples=case_samples)

            else:
                db_case: DbCase = self._update_existing_case(
                    existing_case=case, ticket_id=order._generated_ticket_id
                )

            db_order.cases.append(db_case)
            self.status_db.add_multiple_items_to_store(new_cases)
            self.status_db.add_item_to_store(db_order)
            self.status_db.commit_to_store()
        return new_cases

    @staticmethod
    def _update_case_panel(panels: list[str], case: DbCase) -> None:
        """Update case panels."""
        case.panels = panels

    @staticmethod
    def _append_ticket(ticket_id: str, case: DbCase) -> None:
        """Add a ticket to the case."""
        case.tickets = f"{case.tickets},{ticket_id}"

    @staticmethod
    def _update_action(action: str, case: DbCase) -> None:
        """Update action of a case."""
        case.action = action

    def _create_link(
        self,
        case: DbCase,
        db_sample: DbSample,
        father: DbSample,
        mother: DbSample,
        sample: SampleInCase,
    ) -> CaseSample:
        return self.status_db.relate_sample(
            case=case,
            sample=db_sample,
            status=getattr(sample, "status", None),
            mother=mother,
            father=father,
        )

    def _create_db_sample(
        self,
        case: Case,
        customer: Customer,
        order_name: str,
        ordered: datetime,
        sample: SampleInCase,
        ticket: str,
    ) -> DbSample:
        application_tag = sample.application
        application_version: ApplicationVersion = (
            self.status_db.get_current_application_version_by_tag(tag=application_tag)
        )
        db_sample: DbSample = self.status_db.add_sample(
            application_version=application_version,
            internal_id=sample._generated_lims_id,
            order=order_name,
            ordered=ordered,
            original_ticket=ticket,
            priority=case.priority,
            **sample.model_dump(exclude={"application", "container", "container_name"}),
        )
        db_sample.customer = customer
        self.status_db.add_item_to_store(db_sample)
        return db_sample

    def _create_db_case(
        self,
        case: Case,
        customer: Customer,
        ticket: str,
        workflow: Workflow,
        delivery_type: DataDelivery,
    ) -> DbCase:
        db_case: DbCase = self.status_db.add_case(
            ticket=ticket,
            data_analysis=workflow,
            data_delivery=delivery_type,
            **case.model_dump(exclude={"samples"}),
        )
        db_case.customer = customer
        return db_case

    def _create_db_order(self, order: OrderWithCases) -> DbOrder:
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=order.customer
        )
        return DbOrder(
            customer=customer,
            order_date=datetime.now(),
            ticket_id=order._generated_ticket_id,
        )

    def _update_existing_case(self, existing_case: ExistingCase, ticket_id: int) -> DbCase:
        status_db_case = self.status_db.get_case_by_internal_id(existing_case.internal_id)
        self._append_ticket(ticket_id=str(ticket_id), case=status_db_case)
        self._update_action(action=CaseActions.ANALYZE, case=status_db_case)
        self._update_case_panel(panels=getattr(existing_case, "panels", []), case=status_db_case)
        return status_db_case

    def _create_links(self, case: Case, db_case: DbCase, case_samples: dict[str, DbSample]) -> None:
        """Creates entries in the CaseSample table.
        Input:
        - case: Case, a case within the customer submitted order.
        - db_case: DbCase, Database case entry corresponding to the 'case' parameter.
        - case_samples: dict with keys being sample names in the provided 'case' and values being
        the corresponding database entries in the Sample table."""
        for sample in case.samples:
            if sample.is_new:
                db_sample: DbSample = case_samples.get(sample.name)
            else:
                db_sample: DbSample = self.status_db.get_sample_by_internal_id(sample.internal_id)
            sample_mother_name: str | None = getattr(sample, Pedigree.MOTHER, None)
            db_sample_mother: DbSample | None = case_samples.get(sample_mother_name)
            sample_father_name: str = getattr(sample, Pedigree.FATHER, None)
            db_sample_father: DbSample | None = case_samples.get(sample_father_name)
            case_sample: CaseSample = self._create_link(
                case=db_case,
                db_sample=db_sample,
                father=db_sample_father,
                mother=db_sample_mother,
                sample=sample,
            )
            self.status_db.add_item_to_store(case_sample)

    def _create_db_sample_dict(
        self, case: Case, order: OrderWithCases, customer: Customer
    ) -> dict[str, DbSample]:
        """Constructs a dict containing all the samples in the case. Keys are sample names
        and the values are the database entries for the samples."""
        case_samples: dict[str, DbSample] = {}
        for sample in case.samples:
            if sample.is_new:
                with self.status_db.no_autoflush_context():
                    db_sample: DbSample = self._create_db_sample(
                        case=case,
                        customer=customer,
                        order_name=order.name,
                        ordered=datetime.now(),
                        sample=sample,
                        ticket=str(order._generated_ticket_id),
                    )
            else:
                db_sample: DbSample = self.status_db.get_sample_by_internal_id(sample.internal_id)
            case_samples[db_sample.name] = db_sample
        return case_samples



from cg.constants.priority import Priority, TrailblazerPriority

MAP_TO_TRAILBLAZER_PRIORITY: dict[Priority, TrailblazerPriority] = {
    Priority.research: TrailblazerPriority.LOW,
    Priority.standard: TrailblazerPriority.NORMAL,
    Priority.clinical_trials: TrailblazerPriority.NORMAL,
    Priority.priority: TrailblazerPriority.HIGH,
    Priority.express: TrailblazerPriority.EXPRESS,
}



"""Priority specific constants"""

from enum import IntEnum, StrEnum


class SlurmQos(StrEnum):
    MAINTENANCE = "maintenance"
    LOW = "low"
    NORMAL = "normal"
    HIGH = "high"
    EXPRESS = "express"


class TrailblazerPriority(StrEnum):
    LOW = "low"
    NORMAL = "normal"
    HIGH = "high"
    EXPRESS = "express"


class PriorityTerms(StrEnum):
    EXPRESS: str = "express"
    PRIORITY: str = "priority"
    RESEARCH: str = "research"
    STANDARD: str = "standard"
    CLINICAL_TRIALS: str = "clinical_trials"


class Priority(IntEnum):
    research = 0
    standard = 1
    clinical_trials = 2
    priority = 3
    express = 4

    @classmethod
    def priority_to_slurm_qos(cls) -> dict["Priority", SlurmQos]:
        return {
            Priority.research: SlurmQos.LOW,
            Priority.standard: SlurmQos.NORMAL,
            Priority.priority: SlurmQos.HIGH,
            Priority.express: SlurmQos.EXPRESS,
            Priority.clinical_trials: SlurmQos.NORMAL,
        }


class SlurmAccount(StrEnum):
    PRODUCTION = "production"
    DEVELOPMENT = "development"
    STAGE = "stage"



import logging
from typing import Type

import rich_click as click
from housekeeper.store.models import Archive, File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.archiving import ArchiveLocations
from cg.exc import ArchiveJobFailedError, MissingFilesError, SampleFilesCurrentlyArchivingError
from cg.meta.archive.ddn.ddn_data_flow_client import DDNDataFlowClient
from cg.meta.archive.models import ArchiveHandler, FileAndSample
from cg.models.cg_config import DataFlowConfig
from cg.store.models import Case, Order, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)
ARCHIVE_HANDLERS: dict[str, Type[ArchiveHandler]] = {
    ArchiveLocations.KAROLINSKA_BUCKET: DDNDataFlowClient
}


class SpringArchiveAPI:
    """Class handling the archiving of sample SPRING files to an off-premise location for long
    term storage."""

    def __init__(
        self, housekeeper_api: HousekeeperAPI, status_db: Store, data_flow_config: DataFlowConfig
    ):
        self.housekeeper_api: HousekeeperAPI = housekeeper_api
        self.status_db: Store = status_db
        self.data_flow_config: DataFlowConfig = data_flow_config

    def archive_spring_files_and_add_archives_to_housekeeper(
        self, spring_file_count_limit: int | None
    ) -> None:
        """Archives all non-archived spring files. If a limit is provided, the amount of files archived is limited
        to that amount."""
        if isinstance(spring_file_count_limit, int) and spring_file_count_limit <= 0:
            LOG.warning("Please do not provide a non-positive integer as limit - exiting.")
            return
        for archive_location in ArchiveLocations:
            self.archive_files_to_location(
                archive_location=archive_location, file_limit=spring_file_count_limit
            )

    def archive_files_to_location(self, archive_location: str, file_limit: int | None) -> None:
        """Archives up to spring file count limit number of files to the provided archive location."""
        files_to_archive: list[File] = self.housekeeper_api.get_non_archived_spring_files(
            tags=[archive_location],
            limit=file_limit,
        )
        if files_to_archive:
            files_and_samples_for_location = self.add_samples_to_files(files_to_archive)
            archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](
                self.data_flow_config
            )
            for file_and_sample in files_and_samples_for_location:
                self.archive_file(file_and_sample=file_and_sample, archive_handler=archive_handler)

        else:
            LOG.info(f"No files to archive for location {archive_location}.")

    def archive_file(self, file_and_sample: FileAndSample, archive_handler: ArchiveHandler) -> None:
        job_id: int = archive_handler.archive_file(file_and_sample)
        self.housekeeper_api.add_archives(
            files=[file_and_sample.file],
            archive_task_id=job_id,
        )

    def retrieve_files_from_archive_location(
        self, files_and_samples: list[FileAndSample], archive_location: str
    ) -> None:
        """Retrieves the archived spring files for a list of samples and sets retrieval ids in Housekeeper."""
        archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](self.data_flow_config)
        job_id: int = archive_handler.retrieve_files(files_and_samples)
        LOG.info(f"Retrieval job launched with ID {job_id}")
        self.set_archive_retrieval_task_ids(
            retrieval_task_id=job_id,
            files=[file_and_sample.file for file_and_sample in files_and_samples],
        )

    def get_archived_files_from_sample(self, sample: Sample) -> list[File]:
        """Gets archived spring files from the bundles corresponding to the given list of samples."""
        files: list[File] = self.housekeeper_api.get_archived_files_not_being_retrieved_for_bundle(
            bundle_name=sample.internal_id, tags=[SequencingFileTag.SPRING]
        )
        return files

    def set_archive_retrieval_task_ids(self, retrieval_task_id: int, files: list[File]) -> None:
        for file in files:
            self.housekeeper_api.set_archive_retrieval_task_id(
                file_id=file.id, retrieval_task_id=retrieval_task_id
            )

    def get_sample(self, file: File) -> Sample | None:
        """Fetches the Sample corresponding to a File and logs if a Sample is not found."""
        sample: Sample | None = self.status_db.get_sample_by_internal_id(file.version.bundle.name)
        if not sample:
            LOG.warning(
                f"No sample found in status_db corresponding to sample_id {file.version.bundle.name}."
                f"Skipping archiving for corresponding file {file.path}."
            )
        return sample

    def add_samples_to_files(self, files: list[File]) -> list[FileAndSample]:
        """Fetches the Sample corresponding to each File, instantiates a FileAndSample object and
        adds it to the list which is returned."""
        files_and_samples: list[FileAndSample] = []
        for file in files:
            if sample := self.get_sample(file):
                files_and_samples.append(FileAndSample(file=file, sample=sample))
        return files_and_samples

    def update_statuses_for_ongoing_tasks(self) -> None:
        """Updates any completed jobs with a finished timestamp."""
        self.update_ongoing_archivals()
        self.update_ongoing_retrievals()

    def update_ongoing_archivals(self) -> None:
        ongoing_archivals: list[Archive] = self.housekeeper_api.get_ongoing_archivals()
        archival_ids_per_location: dict[ArchiveLocations, list[int]] = (
            self.sort_archival_ids_on_archive_location(ongoing_archivals)
        )
        for archive_location in ArchiveLocations:
            if archival_ids := archival_ids_per_location.get(archive_location):
                archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](
                    self.data_flow_config
                )
                self.update_archival_jobs_for_archive_location(
                    archive_handler=archive_handler,
                    job_ids=archival_ids,
                )

    def update_ongoing_retrievals(self) -> None:
        ongoing_retrievals: list[Archive] = self.housekeeper_api.get_ongoing_retrievals()
        retrieval_ids_per_location: dict[ArchiveLocations, list[int]] = (
            self.sort_retrieval_ids_on_archive_location(ongoing_retrievals)
        )
        for archive_location in ArchiveLocations:
            if retrieval_ids := retrieval_ids_per_location.get(archive_location):
                archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](
                    self.data_flow_config
                )
                self.update_retrieval_jobs_for_archive_location(
                    archive_handler=archive_handler,
                    job_ids=retrieval_ids,
                )

    def update_archival_jobs_for_archive_location(
        self, archive_handler: ArchiveHandler, job_ids: list[int]
    ) -> None:
        for job_id in job_ids:
            self.update_ongoing_task(
                task_id=job_id, archive_handler=archive_handler, is_archival=True
            )

    def update_retrieval_jobs_for_archive_location(
        self, archive_handler: ArchiveHandler, job_ids: list[int]
    ) -> None:
        for job_id in job_ids:
            self.update_ongoing_task(
                task_id=job_id, archive_handler=archive_handler, is_archival=False
            )

    def update_ongoing_task(
        self, task_id: int, archive_handler: ArchiveHandler, is_archival: bool
    ) -> None:
        """Fetches info on an ongoing job and updates the Archive entry in Housekeeper."""
        try:
            LOG.info(f"Fetching status for job with id {task_id}")
            is_job_done: bool = archive_handler.is_job_done(task_id)
            if is_job_done:
                LOG.info(f"Job with id {task_id} has finished, updating Archive entries.")
                if is_archival:
                    self.housekeeper_api.set_archived_at(task_id)
                else:
                    self.housekeeper_api.set_retrieved_at(task_id)
            else:
                LOG.info(f"Job with id {task_id} has not yet finished.")
        except ArchiveJobFailedError as error:
            LOG.error(error)
            if is_archival:
                LOG.warning(f"Will remove archive entries with archival task ids {task_id}")
                self.housekeeper_api.delete_archives(task_id)
            else:
                LOG.warning("Will set retrieval task id to null.")
                self.housekeeper_api.update_archive_retrieved_at(
                    old_retrieval_job_id=task_id, new_retrieval_job_id=None
                )

    def sort_archival_ids_on_archive_location(
        self, archive_entries: list[Archive]
    ) -> dict[ArchiveLocations, list[int]]:
        """Returns a dictionary with keys being ArchiveLocations and the values being the subset of the given
        archival jobs which should be archived there."""

        jobs_per_location: dict[ArchiveLocations, list[int]] = {}
        jobs_and_locations: set[tuple[int, ArchiveLocations]] = (
            self.get_unique_archival_ids_and_archive_locations(archive_entries)
        )

        for archive_location in ArchiveLocations:
            jobs_per_location[ArchiveLocations(archive_location)] = [
                job_and_location[0]
                for job_and_location in jobs_and_locations
                if job_and_location[1] == archive_location
            ]
        return jobs_per_location

    def get_unique_archival_ids_and_archive_locations(
        self, archive_entries: list[Archive]
    ) -> set[tuple[int, ArchiveLocations]]:
        ids_and_locations: set[tuple[int, ArchiveLocations]] = set()
        for archive in archive_entries:
            if location := self.get_archive_location_from_file(archive.file):
                ids_and_locations.add((archive.archiving_task_id, location))
        return ids_and_locations

    def sort_retrieval_ids_on_archive_location(
        self, archive_entries: list[Archive]
    ) -> dict[ArchiveLocations, list[int]]:
        """Returns a dictionary with keys being ArchiveLocations and the values being the subset of the given
        retrieval jobs which should be archived there."""
        jobs_per_location: dict[ArchiveLocations, list[int]] = {}
        jobs_and_locations: set[tuple[int, ArchiveLocations]] = (
            self.get_unique_retrieval_ids_and_archive_locations(archive_entries)
        )
        for archive_location in ArchiveLocations:
            jobs_per_location[ArchiveLocations(archive_location)] = [
                job_and_location[0]
                for job_and_location in jobs_and_locations
                if job_and_location[1] == archive_location
            ]
        return jobs_per_location

    def get_unique_retrieval_ids_and_archive_locations(
        self, archive_entries: list[Archive]
    ) -> set[tuple[int, ArchiveLocations]]:
        ids_and_locations: set[tuple[int, ArchiveLocations]] = set()
        for archive in archive_entries:
            if location := self.get_archive_location_from_file(archive.file):
                ids_and_locations.add((archive.retrieval_task_id, location))
        return ids_and_locations

    @staticmethod
    def is_file_archived(file: File) -> bool:
        return file.archive and file.archive.archived_at

    @staticmethod
    def get_archive_location_from_file(file: File) -> ArchiveLocations | None:
        for tag_name in [tag.name for tag in file.tags]:
            if tag_name in iter(ArchiveLocations):
                LOG.debug(f"Found archive location {tag_name}")
                return tag_name
        LOG.warning("No archive location in the file tags")
        return None

    def delete_file_from_archive_location(
        self, file_and_sample: FileAndSample, archive_location: ArchiveLocations
    ) -> None:
        archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](self.data_flow_config)
        archive_handler.delete_file(file_and_sample)

    def delete_file(self, file_path: str, dry_run: bool = False) -> None:
        """Deletes the specified file where it is archived and deletes the Housekeeper record.
        Raises:
            Click.Abort if yes is not specified or the user does not confirm the deletion."""
        file: File = self.housekeeper_api.files(path=file_path).first()
        if not self.is_file_archived(file):
            LOG.warning(f"No archived file found for file {file_path} - exiting")
            return
        archive_location: ArchiveLocations | None = self.get_archive_location_from_file(file)
        if not archive_location:
            LOG.warning("No archive location could be determined - exiting")
            return
        if dry_run:
            click.echo(f"Would have deleted file {file_path} from {archive_location}.")
            return
        file_and_sample: FileAndSample = self.add_samples_to_files([file])[0]
        self.delete_file_from_archive_location(
            file_and_sample=file_and_sample, archive_location=archive_location
        )
        self.housekeeper_api.delete_file(file.id)

    def retrieve_spring_files_for_order(self, id_: int, is_order_id: bool) -> None:
        if is_order_id:
            order: Order = self.status_db.get_order_by_id(id_)
        else:
            order = self.status_db.get_order_by_ticket_id(id_)
        for case in order.cases:
            self.retrieve_spring_files_for_case(case.internal_id)

    def retrieve_spring_files_for_case(self, case_id: str) -> None:
        """Submits jobs to retrieve any archived files belonging to the given case, and updates the Archive entries
        with the retrieval job id."""
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        for sample in case.samples:
            try:
                self._retrieve_spring_files_for_sample(sample)
            except MissingFilesError as error:
                LOG.warning(str(error))
                continue
            except SampleFilesCurrentlyArchivingError as error:
                LOG.warning(str(error))
                continue

    def retrieve_spring_files_for_sample(self, sample_id: str) -> None:
        sample: Sample = self.status_db.get_sample_by_internal_id(sample_id)
        self._retrieve_spring_files_for_sample(sample)

    def _retrieve_spring_files_for_sample(self, sample: Sample) -> None:
        files_to_retrieve: list[File] = self.get_archived_files_from_sample(sample)
        if not files_to_retrieve:
            raise MissingFilesError(
                f"No archived Spring files found for sample {sample.internal_id}."
            )
        if not self._are_all_files_archived(files_to_retrieve):
            raise SampleFilesCurrentlyArchivingError(
                f"Not all Spring files for sample {sample.internal_id} are archived - cannot retrieve files."
            )
        files_and_samples: list[FileAndSample] = self.add_samples_to_files(files_to_retrieve)
        self.retrieve_files_from_archive_location(
            files_and_samples=files_and_samples, archive_location=sample.archive_location
        )

    def _are_all_files_archived(self, files: list[File]) -> bool:
        return all(self.is_file_archived(file) for file in files)



import logging

from pydantic import TypeAdapter

from cg.apps.demultiplex.sample_sheet.sample_models import IlluminaSampleIndexSetting
from cg.constants.demultiplexing import SampleSheetBcl2FastqSections, SampleSheetBCLConvertSections
from cg.exc import SampleSheetContentError, SampleSheetFormatError

LOG = logging.getLogger(__name__)


def validate_samples_are_unique(samples: list[IlluminaSampleIndexSetting]) -> None:
    """Validate that each sample only exists once."""
    sample_ids: set = set()
    for sample in samples:
        sample_id: str = sample.sample_id.split("_")[0]
        if sample_id in sample_ids:
            message: str = f"Sample {sample.sample_id} exists multiple times in sample sheet"
            LOG.error(message)
            raise SampleSheetContentError(message)
        sample_ids.add(sample_id)


def validate_samples_unique_per_lane(samples: list[IlluminaSampleIndexSetting]) -> None:
    """Validate that each sample only exists once per lane in a sample sheet."""
    sample_by_lane: dict[int, list[IlluminaSampleIndexSetting]] = get_samples_by_lane(samples)
    for lane, lane_samples in sample_by_lane.items():
        LOG.debug(f"Validate that samples are unique in lane: {lane}")
        validate_samples_are_unique(samples=lane_samples)


def get_raw_samples_from_content(sample_sheet_content: list[list[str]]) -> list[dict[str, str]]:
    """Return the samples in a sample sheet as a list of dictionaries."""
    header: list[str] = []
    raw_samples: list[dict[str, str]] = []

    for line in sample_sheet_content:
        # Skip lines that are too short to contain samples
        if len(line) <= 4:
            continue
        if line[0] in [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBCLConvertSections.Data.LANE.value,
        ]:
            header = line
            continue
        if not header:
            continue
        raw_samples.append(dict(zip(header, line)))
    if not header:
        message = "Could not find header in sample sheet"
        LOG.warning(message)
        raise SampleSheetFormatError(message)
    if not raw_samples:
        message = "Could not find any samples in sample sheet"
        LOG.warning(message)
        raise SampleSheetFormatError(message)
    return raw_samples


def get_samples_by_lane(
    samples: list[IlluminaSampleIndexSetting],
) -> dict[int, list[IlluminaSampleIndexSetting]]:
    """Group and return samples by lane."""
    LOG.debug("Order samples by lane")
    sample_by_lane: dict[int, list[IlluminaSampleIndexSetting]] = {}
    for sample in samples:
        if sample.lane not in sample_by_lane:
            sample_by_lane[sample.lane] = []
        sample_by_lane[sample.lane].append(sample)
    return sample_by_lane


def get_samples_from_content(
    sample_sheet_content: list[list[str]],
) -> list[IlluminaSampleIndexSetting]:
    """
    Return the samples in a sample sheet as a list of IlluminaIndexSettings objects.
    Raises:
        ValidationError: if the samples do not have the correct attributes based on their model.
    """
    raw_samples: list[dict[str, str]] = get_raw_samples_from_content(
        sample_sheet_content=sample_sheet_content
    )
    adapter = TypeAdapter(list[IlluminaSampleIndexSetting])
    return adapter.validate_python(raw_samples)



from datetime import datetime

from cg.server.dto.samples.requests import SamplesRequest
from cg.server.dto.samples.samples_response import SamplesResponse
from cg.services.web_services.sample.dto_mappers import create_sample_dto
from cg.store.models import Sample


def get_cancel_comment(user_name: str) -> str:
    date: str = datetime.now().strftime("%Y-%m-%d")
    return f"Cancelled {date} by {user_name}"


def get_confirmation_message(sample_ids: list[str], case_ids: list[str]) -> str:
    message = f"Cancelled {len(sample_ids)} samples. "
    if case_ids:
        cases = ", ".join(case_ids)
        message += f"Found {len(case_ids)} cases with additional samples: {cases}."
    else:
        message += "No case contained additional samples."
    return message


def get_start_and_finish_indexes_from_request(request: SamplesRequest) -> tuple[int, int]:
    """
    Return the start and finish indexes to slice a sample list given the pagination information
    in the sample request.
    """
    page_size = request.page_size or 50
    page = request.page or 1
    start = (page - 1) * page_size
    finish = start + page_size
    return start, finish


def create_samples_response(samples: list[Sample]) -> SamplesResponse:
    sample_dtos = []
    for sample in samples:
        sample_dtos.append(create_sample_dto(sample))
    return SamplesResponse(samples=sample_dtos, total=len(samples))



from flask import Blueprint, jsonify

from cg.server.endpoints.sequencing_metrics.error_handler import handle_endpoint_errors
from cg.server.endpoints.utils import before_request
from cg.server.ext import sample_run_metrics_service
from cg.services.sample_run_metrics_service.dtos import IlluminaSequencingMetricsDTO

FLOW_CELLS_BLUEPRINT = Blueprint("flowcells", __name__, url_prefix="/api/v1")
FLOW_CELLS_BLUEPRINT.before_request(before_request)


@FLOW_CELLS_BLUEPRINT.route("/flowcells/<flow_cell_name>/sequencing_metrics", methods=["GET"])
@handle_endpoint_errors
def get_sequencing_metrics(flow_cell_name: str):
    metrics: list[IlluminaSequencingMetricsDTO] = sample_run_metrics_service.get_illumina_metrics(
        flow_cell_name
    )
    return jsonify([metric.model_dump() for metric in metrics])



from typing import Any


def has_internal_id(v: Any) -> str:
    if isinstance(v, dict):
        return "existing" if v.get("internal_id") else "new"
    return "existing" if getattr(v, "internal_id", None) else "new"



"""Class to hold information about scout load config"""

from datetime import datetime

from pydantic import AfterValidator, BaseModel, ConfigDict
from typing_extensions import Annotated, Literal

from cg.constants.scout import UploadTrack
from cg.models.scout.validators import field_not_none


class ChromographImages(BaseModel):
    autozygous: str | None = None
    coverage: str | None = None
    upd_regions: str | None = None
    upd_sites: str | None = None


class Eklipse(BaseModel):
    title: str | None = None
    description: str | None = None
    width: str | None = None
    height: str | None = None
    path: str | None = None


class CaseImages(BaseModel):
    eKLIPse: list[Eklipse] | None = None


class CustomImages(BaseModel):
    case_images: CaseImages | None = None


class Reviewer(BaseModel):
    alignment: str | None = None
    alignment_index: str | None = None
    vcf: str | None = None
    catalog: str | None = None
    trgt: bool | None = None


class ScoutIndividual(BaseModel):
    alignment_path: str | None = None
    rna_alignment_path: str | None = None
    analysis_type: Annotated[
        Literal[
            "external",
            "mixed",
            "panel",
            "panel-umi",
            "unknown",
            "wes",
            "wgs",
            "wts",
        ],
        AfterValidator(field_not_none),
    ] = None
    capture_kit: str | None = None
    confirmed_parent: bool | None = None
    confirmed_sex: bool | None = None
    father: str | None = None
    mother: str | None = None
    phenotype: str | None = None
    sample_id: Annotated[str | None, AfterValidator(field_not_none)] = None
    sample_name: str | None = None
    sex: Annotated[str | None, AfterValidator(field_not_none)] = None
    subject_id: str | None = None
    tissue_type: str | None = None
    model_config = ConfigDict(validate_assignment=True)


class ScoutMipIndividual(ScoutIndividual):
    mt_bam: str | None = None
    chromograph_images: ChromographImages = ChromographImages()
    reviewer: Reviewer = Reviewer()
    rhocall_bed: str | None = None
    rhocall_wig: str | None = None
    tiddit_coverage_wig: str | None = None
    upd_regions_bed: str | None = None
    upd_sites_bed: str | None = None
    vcf2cytosure: str | None = None
    mitodel_file: str | None = None


class ScoutNalloIndividual(ScoutIndividual):
    assembly_alignment_path: str | None = None
    d4_file: str | None = None
    minor_allele_frequency_wig: str | None = None
    mt_bam: str | None = None
    paraphase_alignment_path: str | None = None
    phase_blocks: str | None = None
    reviewer: Reviewer = Reviewer()
    tiddit_coverage_wig: str | None = None


class ScoutRarediseaseIndividual(ScoutIndividual):
    mt_bam: str | None = None
    chromograph_images: ChromographImages = ChromographImages()
    rhocall_bed: str | None = None
    rhocall_wig: str | None = None
    tiddit_coverage_wig: str | None = None
    upd_regions_bed: str | None = None
    upd_sites_bed: str | None = None
    vcf2cytosure: str | None = None
    mitodel_file: str | None = None
    reviewer: Reviewer = Reviewer()
    d4_file: str | None = None


class ScoutCancerIndividual(ScoutIndividual):
    tumor_type: str | None = None
    tmb: str | None = None
    msi: str | None = None
    tumor_purity: float = 0
    vcf2cytosure: str | None = None


class ScoutLoadConfig(BaseModel):
    owner: Annotated[str | None, AfterValidator(field_not_none)] = None
    family: Annotated[str | None, AfterValidator(field_not_none)] = None
    family_name: str | None = None
    status: str | None = None
    synopsis: str | None = None
    phenotype_terms: list[str] | None = None
    phenotype_groups: list[str] | None = None
    gene_panels: list[str] | None = None
    default_gene_panels: list[str] = []
    cohorts: list[str] | None = None
    human_genome_build: str = None
    rank_model_version: str | None = None
    rank_score_threshold: int = None
    sv_rank_model_version: str | None = None
    analysis_date: datetime | None = None
    samples: list[ScoutIndividual] = []
    customer_images: CustomImages | None = None
    delivery_report: str | None = None
    coverage_qc_report: str | None = None
    cnv_report: str | None = None
    multiqc: str | None = None
    track: Literal[UploadTrack.RARE_DISEASE.value, UploadTrack.CANCER.value] = (
        UploadTrack.RARE_DISEASE.value
    )
    model_config = ConfigDict(validate_assignment=True)


class BalsamicLoadConfig(ScoutLoadConfig):
    madeline: str | None = None
    vcf_cancer: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_cancer_sv: str | None = None
    vcf_cancer_research: str | None = None
    vcf_cancer_sv_research: str | None = None
    samples: list[ScoutCancerIndividual] = []


class BalsamicUmiLoadConfig(BalsamicLoadConfig):
    pass


class MipLoadConfig(ScoutLoadConfig):
    chromograph_image_files: list[str] | None = None
    chromograph_prefixes: list[str] | None = None
    madeline: str | None = None
    peddy_check: str | None = None
    peddy_ped: str | None = None
    peddy_sex: str | None = None
    samples: list[ScoutMipIndividual] = []
    smn_tsv: str | None = None
    vcf_mei: str | None = None
    vcf_mei_research: str | None = None
    vcf_snv: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_snv_research: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_str: str | None = None
    vcf_sv: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_sv_research: Annotated[str | None, AfterValidator(field_not_none)] = None


class NalloLoadConfig(ScoutLoadConfig):
    madeline: str | None = None
    peddy_check: str | None = None
    peddy_ped: str | None = None
    peddy_sex: str | None = None
    samples: list[ScoutNalloIndividual] = []
    somalier_samples: str | None = None
    somalier_pairs: str | None = None
    vcf_snv: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_snv_research: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_sv: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_sv_research: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_str: str | None = None


class RarediseaseLoadConfig(ScoutLoadConfig):
    madeline: str | None = None
    peddy_check: str | None = None
    peddy_ped: str | None = None
    peddy_sex: str | None = None
    samples: list[ScoutRarediseaseIndividual] = []
    custom_images: CustomImages | None = None
    smn_tsv: str | None = None
    vcf_mei: str | None = None
    vcf_mei_research: str | None = None
    vcf_snv: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_snv_research: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_snv_mt: str | None = None
    vcf_snv_research_mt: str | None = None
    vcf_sv: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_sv_research: Annotated[str | None, AfterValidator(field_not_none)] = None
    vcf_str: str | None = None


class RnafusionLoadConfig(ScoutLoadConfig):
    multiqc_rna: str | None = None
    gene_fusion: str | None = None
    gene_fusion_report_research: str | None = None
    RNAfusion_inspector: str | None = None
    RNAfusion_inspector_research: str | None = None
    RNAfusion_report: str | None = None
    RNAfusion_report_research: str | None = None
    samples: list[ScoutCancerIndividual] = []
    vcf_fusion: str | None = None



"""CLI support to create config and/or start BALSAMIC """

import logging
import traceback

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.balsamic.base import report_deliver, store, store_available, store_housekeeper
from cg.cli.workflow.balsamic.options import OPTION_PANEL_BED, OPTION_WORKFLOW_PROFILE
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, link, resolve_compression
from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Sex
from cg.constants.cli_options import DRY_RUN, LIMIT
from cg.constants.constants import GenomeVersion
from cg.constants.priority import SlurmQos
from cg.exc import AnalysisNotReadyError, CgError
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case

LOG = logging.getLogger(__name__)

OPTION_GENDER = click.option(
    "--gender",
    type=click.Choice([Sex.FEMALE, Sex.MALE]),
    required=False,
    help="Case associated gender. Set this option to override the one selected by the customer in StatusDB.",
)
OPTION_GENOME_VERSION = click.option(
    "--genome-version",
    show_default=True,
    default=GenomeVersion.HG19,
    type=click.Choice([GenomeVersion.HG19, GenomeVersion.HG38, GenomeVersion.CANFAM3]),
    help="Type and build version of the reference genome. Set this option to override the default.",
)
OPTION_QOS = click.option(
    "-qos",
    "--slurm-quality-of-service",
    type=click.Choice([SlurmQos.LOW, SlurmQos.NORMAL, SlurmQos.HIGH, SlurmQos.EXPRESS]),
    help="Job priority in SLURM. Setting this option will override the StatusDB case priority.",
)
OPTION_PON_CNN = click.option(
    "--pon-cnn",
    type=click.Path(exists=True),
    required=False,
    help="Panel of normal reference (.cnn) for CNVkit",
)
OPTION_OBSERVATIONS = click.option(
    "--observations",
    type=click.Path(exists=True),
    multiple=True,
    required=False,
    help="VCF paths of clinical and/or cancer SNVs and SVs observations (WHOLE_GENOME_SEQUENCING analysis only). Set this option to "
    "override the latest Loqusdb dump files.",
)
OPTION_CACHE_VERSION = click.option(
    "--cache-version",
    type=click.STRING,
    required=False,
    help="Cache version to be used for init or analysis. Use 'develop' or 'X.X.X'.",
)


@click.group(
    "balsamic-umi",
    invoke_without_command=True,
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.pass_context
def balsamic_umi(context: click.Context):
    """Cancer analysis workflow"""
    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return None
    config = context.obj
    context.obj.meta_apis["analysis_api"] = BalsamicUmiAnalysisAPI(config=config)


@balsamic_umi.command("config-case")
@ARGUMENT_CASE_ID
@OPTION_GENDER
@OPTION_GENOME_VERSION
@OPTION_PANEL_BED
@OPTION_PON_CNN
@OPTION_OBSERVATIONS
@OPTION_CACHE_VERSION
@DRY_RUN
@click.pass_obj
def config_case(
    context: CGConfig,
    case_id: str,
    gender: str,
    genome_version: str,
    panel_bed: str,
    pon_cnn: click.Path,
    observations: list[click.Path],
    cache_version: str,
    dry_run: bool,
):
    """Create config file for BALSAMIC UMI analysis for a given CASE_ID."""

    analysis_api: BalsamicUmiAnalysisAPI = context.meta_apis["analysis_api"]
    try:
        LOG.info(f"Creating config file for {case_id}.")
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.config_case(
            case_id=case_id,
            gender=gender,
            genome_version=genome_version,
            panel_bed=panel_bed,
            pon_cnn=pon_cnn,
            observations=observations,
            cache_version=cache_version,
            dry_run=dry_run,
        )
    except CgError as error:
        error_info = f"Error: {type(error).__name__}: {str(error)}\n{traceback.format_exc()}"
        LOG.error(f"Could not create config: {error_info}")
        raise click.Abort()
    except Exception as error:
        error_info = f"Error: {type(error).__name__}: {str(error)}\n{traceback.format_exc()}"
        LOG.error(f"Could not create config: {error_info}")
        raise click.Abort()


@balsamic_umi.command("run")
@ARGUMENT_CASE_ID
@OPTION_WORKFLOW_PROFILE
@DRY_RUN
@OPTION_QOS
@click.pass_obj
def run(
    context: CGConfig,
    case_id: str,
    workflow_profile: click.Path,
    slurm_quality_of_service: str,
    dry_run: bool,
):
    """Run Balsamic UMI analysis for given CASE ID"""
    analysis_api: BalsamicUmiAnalysisAPI = context.meta_apis["analysis_api"]
    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.verify_case_config_file_exists(case_id=case_id, dry_run=dry_run)
        analysis_api.check_analysis_ongoing(case_id)
        analysis_api.run_analysis(
            case_id=case_id,
            workflow_profile=workflow_profile,
            slurm_quality_of_service=slurm_quality_of_service,
            dry_run=dry_run,
        )
        if dry_run:
            return
        analysis_api.on_analysis_started(case_id)
    except Exception as error:
        error_info = f"Error: {type(error).__name__}: {str(error)}\n{traceback.format_exc()}"
        LOG.error(f"Could not run analysis: {error_info}")
        raise click.Abort()


@balsamic_umi.command("start")
@ARGUMENT_CASE_ID
@OPTION_GENDER
@OPTION_GENOME_VERSION
@OPTION_QOS
@DRY_RUN
@OPTION_PANEL_BED
@OPTION_PON_CNN
@OPTION_CACHE_VERSION
@OPTION_OBSERVATIONS
@OPTION_WORKFLOW_PROFILE
@click.pass_context
def start(
    context: click.Context,
    case_id: str,
    gender: str,
    genome_version: str,
    cache_version: str,
    panel_bed: str,
    pon_cnn: str,
    observations: list[click.Path],
    slurm_quality_of_service: str,
    workflow_profile: click.Path,
    dry_run: bool,
):
    """Start full workflow for case ID."""
    analysis_api: BalsamicUmiAnalysisAPI = context.obj.meta_apis["analysis_api"]
    analysis_api.prepare_fastq_files(case_id=case_id, dry_run=dry_run)
    LOG.info(f"Starting analysis for {case_id}")
    context.invoke(link, case_id=case_id, dry_run=dry_run)
    context.invoke(
        config_case,
        case_id=case_id,
        gender=gender,
        genome_version=genome_version,
        cache_version=cache_version,
        panel_bed=panel_bed,
        pon_cnn=pon_cnn,
        observations=observations,
        dry_run=dry_run,
    )
    context.invoke(
        run,
        case_id=case_id,
        workflow_profile=workflow_profile,
        slurm_quality_of_service=slurm_quality_of_service,
        dry_run=dry_run,
    )


@balsamic_umi.command("start-available")
@DRY_RUN
@LIMIT
@click.pass_context
def start_available(context: click.Context, dry_run: bool = False, limit: int | None = None):
    """Start full workflow for all cases ready for analysis"""

    analysis_api: BalsamicUmiAnalysisAPI = context.obj.meta_apis["analysis_api"]

    cases: list[Case] = analysis_api.get_cases_to_analyze(limit=limit)
    LOG.info(f"Starting {len(cases)} available Balsamic UMI cases")

    exit_code: int = EXIT_SUCCESS
    for case in cases:
        try:
            context.invoke(start, case_id=case.internal_id, dry_run=dry_run)
        except AnalysisNotReadyError as error:
            LOG.error(error)
        except CgError as error:
            LOG.error(error)
            exit_code = EXIT_FAIL
        except Exception as error:
            LOG.error(f"Unspecified error occurred: {error}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort


balsamic_umi.add_command(resolve_compression)
balsamic_umi.add_command(link)
balsamic_umi.add_command(report_deliver)
balsamic_umi.add_command(store_housekeeper)
balsamic_umi.add_command(store)
balsamic_umi.add_command(store_available)



from pydantic import BaseModel, ConfigDict

from cg.models.nallo.nallo import NalloQCMetrics
from cg.models.raredisease.raredisease import RarediseaseQCMetrics
from cg.models.rnafusion.rnafusion import RnafusionQCMetrics
from cg.models.taxprofiler.taxprofiler import TaxprofilerQCMetrics
from cg.models.tomte.tomte import TomteQCMetrics


class AnalysisModel(BaseModel):
    """Metadata analysis model"""

    model_config = ConfigDict(arbitrary_types_allowed=True)


class NextflowAnalysis(AnalysisModel):
    """Nextflow's analysis results model."""

    sample_metrics: (
        dict[str, RarediseaseQCMetrics]
        | dict[str, RnafusionQCMetrics]
        | dict[str, TaxprofilerQCMetrics]
        | dict[str, TomteQCMetrics]
        | dict[str, NalloQCMetrics]
    )



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.microsalt.constants import MicrosaltDeliveryType
from cg.services.orders.validation.order_types.microsalt.models.sample import MicrosaltSample


class MicrosaltOrder(OrderWithSamples):
    delivery_type: MicrosaltDeliveryType
    samples: list[MicrosaltSample]

    @property
    def enumerated_samples(self) -> enumerate[MicrosaltSample]:
        return enumerate(self.samples)



from cg.server.endpoints.sequencing_run.dtos import (
    PacbioSequencingRunDTO,
    PacbioSequencingRunsResponse,
)
from cg.store.models import PacbioSequencingRun
from cg.store.store import Store


class PacbioSequencingRunsService:
    def __init__(self, store: Store):
        self.store = store

    def get_sequencing_runs_by_name(self, run_name: str) -> PacbioSequencingRunsResponse:
        runs: list[PacbioSequencingRunDTO] = []
        db_runs: list[PacbioSequencingRun] = self.store.get_pacbio_sequencing_runs_by_run_name(
            run_name
        )
        for db_run in db_runs:
            run_dict = db_run.to_dict()
            run_dict["internal_id"] = db_run.device.internal_id
            run = PacbioSequencingRunDTO.model_validate(run_dict)
            runs.append(run)
        return PacbioSequencingRunsResponse(runs=runs)



import logging
from pathlib import Path

from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.constants import PIPELINES_USING_PARTIAL_ANALYSES, CaseActions
from cg.exc import AnalysisNotReadyError, DecompressionCouldNotStartError
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.compress import CompressAPI, files
from cg.models.compression_data import CaseCompressionData, CompressionData, SampleCompressionData
from cg.services.analysis_starter.input_fetcher.input_fetcher import InputFetcher
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class FastqFetcher(InputFetcher):

    def __init__(
        self,
        compress_api: CompressAPI,
        housekeeper_api: HousekeeperAPI,
        spring_archive_api: SpringArchiveAPI,
        status_db: Store,
    ):
        self.status_db = status_db
        self.housekeeper_api = housekeeper_api
        self.compress_api = compress_api
        self.spring_archive_api = spring_archive_api

    def ensure_files_are_ready(self, case_id: str) -> None:
        """
        Ensures FASTQ files are ready to be analysed. Executes the following processes if needed:
        1. Retrieval of flow cells via PDC.
        2. Retrieval of Spring files via DDN.
        3. Decompression of Spring files.
        4. Adds decompressed FASTQ files to Housekeeper.
        Raises:
            AnalysisNotReadyError if the FASTQ files are not ready to be analysed.
        """
        self._ensure_files_are_present(case_id)
        self._resolve_decompression(case_id)
        if not self._are_fastq_files_ready_for_analysis(case_id):
            raise AnalysisNotReadyError(
                f"FASTQ files needed to start case {case_id} are not present"
            )
        LOG.info(f"All FASTQ files are ready for analysing case {case_id}")

    def _ensure_files_are_present(self, case_id: str) -> None:
        """Checks if any Illumina runs need to be retrieved and queries PDC if that is the case.
        Also checks if any spring files are archived and submits a job to retrieve any which are."""
        self._ensure_illumina_run_on_disk(case_id)
        if not self._are_all_spring_files_present(case_id):
            LOG.warning(f"Files are archived for case {case_id}")
            self.spring_archive_api.retrieve_spring_files_for_case(case_id)

    def _ensure_illumina_run_on_disk(self, case_id: str) -> None:
        """Check if Illumina sequencing runs are on disk for given case. Downsampled and external
        cases are disregarded."""
        if not self._is_illumina_run_check_applicable(case_id):
            LOG.info(
                "Illumina run check is not applicable - "
                "the case is either down sampled or external."
            )
            return
        if not self.status_db.are_all_illumina_runs_on_disk(case_id):
            self.status_db.request_sequencing_runs_for_case(case_id)

    def _are_all_spring_files_present(self, case_id: str) -> bool:
        """Return True if no Spring files for the case are archived."""
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        for sample in case.samples:
            if (
                files := self.housekeeper_api.get_archived_files_for_bundle(
                    bundle_name=sample.internal_id, tags=[SequencingFileTag.SPRING]
                )
            ) and not all(file.archive.retrieved_at for file in files):
                return False
        return True

    def _resolve_decompression(self, case_id: str) -> None:
        """
        Decompresses a case if needed and adds new fastq files to Housekeeper.
        """
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        case_compression_data: CaseCompressionData = self.compress_api.get_case_compression_data(
            case
        )
        if case_compression_data.is_spring_decompression_needed():
            LOG.info(f"The analysis for {case_id} could not start, decompression is needed")
            if not case_compression_data.can_at_least_one_sample_be_decompressed():
                self._set_to_analyze_if_decompressing(case_compression_data)
            else:
                self._decompress_case(case_id)
        elif case_compression_data.is_spring_decompression_running():
            self.status_db.update_case_action(case_internal_id=case_id, action=CaseActions.ANALYZE)
            return

        self._add_decompressed_fastq_files_to_housekeeper(case_id)

    def _decompress_case(self, case_id: str) -> None:
        """Decompress all possible fastq files for all samples in a case"""
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        try:
            self.compress_api.decompress_case(case)
        except DecompressionCouldNotStartError:
            LOG.warning(f"Decompression failed to start for {case_id}")
            return
        self.status_db.update_case_action(case_internal_id=case_id, action=CaseActions.ANALYZE)
        LOG.info(f"Decompression started for {case_id}")

    def _are_fastq_files_ready_for_analysis(self, case_id: str) -> bool:
        """Returns True if no files need to be retrieved from an external location and if all Spring files are
        decompressed."""
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        case_compression_data: CaseCompressionData = self.compress_api.get_case_compression_data(
            case
        )
        if self._does_any_file_need_to_be_retrieved(case_id):
            return False
        if (
            case_compression_data.is_spring_decompression_needed()
            or case_compression_data.is_spring_decompression_running()
        ):
            LOG.warning(f"Case {case_id} is not ready - not all files are decompressed.")
            return False
        return True

    def _does_any_file_need_to_be_retrieved(self, case_id: str) -> bool:
        """Checks whether we need to retrieve files from an external data location."""
        if self._is_illumina_run_check_applicable(
            case_id
        ) and not self.status_db.are_all_illumina_runs_on_disk(case_id):
            LOG.warning(f"Case {case_id} is not ready - not all Illumina runs present on disk.")
            return True
        else:
            if not self._are_all_spring_files_present(case_id):
                LOG.warning(f"Case {case_id} is not ready - some files are archived.")
                return True
        return False

    def _is_illumina_run_check_applicable(self, case_id) -> bool:
        """Returns true if the case is neither down sampled nor external."""
        return not (
            self.status_db.is_case_down_sampled(case_id) or self.status_db.is_case_external(case_id)
        )

    def _set_to_analyze_if_decompressing(self, case_compression_data: CaseCompressionData) -> None:
        """Sets a case's action to 'analyze' if it is currently being decompressed."""
        case_id: str = case_compression_data.case_id
        if case_compression_data.is_spring_decompression_running():
            LOG.info(
                f"Decompression is running for {case_id}, analysis will be started when decompression is done"
            )
            self.status_db.update_case_action(case_internal_id=case_id, action=CaseActions.ANALYZE)

    @staticmethod
    def _should_skip_sample(case: Case, sample: Sample) -> bool:
        """
        For some workflows, we want to start a partial analysis disregarding the samples with no reads.
        This method returns true if we should skip the sample.
        """
        return case.data_analysis in PIPELINES_USING_PARTIAL_ANALYSES and not sample.has_reads

    def _add_decompressed_fastq_files_to_housekeeper(self, case_id: str) -> None:
        """Adds decompressed FASTQ files to Housekeeper for a case, if there are any."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        for sample in case.samples:
            if self._should_skip_sample(case=case, sample=sample):
                LOG.warning(f"Skipping sample {sample.internal_id} - it has no reads.")
                continue
            self._add_decompressed_sample(sample=sample)

    def _add_decompressed_sample(self, sample: Sample) -> None:
        """Adds decompressed FASTQ files to Housekeeper for a sample, if there are any."""
        sample_id = sample.internal_id
        compression_data: SampleCompressionData = self.compress_api.get_sample_compression_data(
            sample_id
        )
        if not self._are_all_fastqs_in_housekeeper(compression_data):
            self.compress_api.add_decompressed_fastq(sample)

    def _are_all_fastqs_in_housekeeper(
        self, sample_compression_data: SampleCompressionData
    ) -> bool:
        """Returns true if all Spring files are decompressed for the sample."""
        sample_id: str = sample_compression_data.sample_id
        version: Version = self.housekeeper_api.get_latest_bundle_version(sample_id)
        fastq_files = files.get_hk_files_dict(tags=[SequencingFileTag.FASTQ], version_obj=version)
        return all(
            self._is_fastq_pair_in_housekeeper(compression=compression, fastq_files=fastq_files)
            for compression in sample_compression_data.compression_objects
        )

    @staticmethod
    def _is_fastq_pair_in_housekeeper(
        compression: CompressionData, fastq_files: dict[Path, File]
    ) -> bool:
        return compression.fastq_first in fastq_files and compression.fastq_second in fastq_files



from http import HTTPStatus

from flask import Blueprint, jsonify

from cg.server.endpoints.utils import before_request
from cg.services.orders.validation.index_sequences import INDEX_SEQUENCES

INDEX_SEQUENCES_BLUEPRINT = Blueprint("index_sequences", __name__, url_prefix="/api/v1")
INDEX_SEQUENCES_BLUEPRINT.before_request(before_request)


@INDEX_SEQUENCES_BLUEPRINT.route("index_sequences", methods=["GET"])
def get_index_sequences():
    return jsonify(INDEX_SEQUENCES), HTTPStatus.OK



from datetime import datetime

from pydantic import BaseModel


class PacbioSequencingRunDTO(BaseModel):
    barcoded_hifi_mean_read_length: int
    barcoded_hifi_reads: int
    barcoded_hifi_reads_percentage: float
    completed_at: datetime
    hifi_mean_read_length: int
    hifi_median_read_quality: str
    hifi_reads: int
    hifi_yield: int
    internal_id: str
    movie_name: str
    p0_percent: float
    p1_percent: float
    p2_percent: float
    percent_reads_passing_q30: float
    plate: int
    run_name: str
    started_at: datetime
    well: str


class PacbioSequencingRunsResponse(BaseModel):
    runs: list[PacbioSequencingRunDTO]



from enum import StrEnum

from cg.constants import DataDelivery


class BalsamicUmiDeliveryType(StrEnum):
    ANALYSIS = DataDelivery.ANALYSIS_FILES
    ANALYSIS_SCOUT = DataDelivery.ANALYSIS_SCOUT
    FASTQ_ANALYSIS = DataDelivery.FASTQ_ANALYSIS
    FASTQ_ANALYSIS_SCOUT = DataDelivery.FASTQ_ANALYSIS_SCOUT
    FASTQ_SCOUT = DataDelivery.FASTQ_SCOUT
    NO_DELIVERY = DataDelivery.NO_DELIVERY
    SCOUT = DataDelivery.SCOUT



from flask import Blueprint, jsonify

from cg.server.endpoints.error_handler import handle_missing_entries
from cg.server.endpoints.sequencing_run.dtos import PacbioSequencingRunsResponse
from cg.server.endpoints.utils import before_request
from cg.server.ext import pacbio_sequencing_runs_service

PACBIO_SEQUENCING_RUN_BLUEPRINT = Blueprint(
    "pacbio_sequencing_run", __name__, url_prefix="/api/v1/pacbio_sequencing_run"
)
PACBIO_SEQUENCING_RUN_BLUEPRINT.before_request(before_request)


@PACBIO_SEQUENCING_RUN_BLUEPRINT.route("/<run_name>", methods=["GET"])
@handle_missing_entries
def get_sequencing_runs(run_name: str):
    response: PacbioSequencingRunsResponse = (
        pacbio_sequencing_runs_service.get_sequencing_runs_by_name(run_name)
    )
    return jsonify(response.model_dump())



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_case import ExistingCase
from cg.services.orders.validation.order_types.balsamic.models.order import BalsamicOrder
from cg.services.orders.validation.order_types.balsamic_umi.constants import BalsamicUmiDeliveryType
from cg.services.orders.validation.order_types.balsamic_umi.models.case import BalsamicUmiCase

NewCase = Annotated[BalsamicUmiCase, Tag("new")]
OldCase = Annotated[ExistingCase, Tag("existing")]


class BalsamicUmiOrder(BalsamicOrder):
    cases: list[Annotated[NewCase | OldCase, Discriminator(has_internal_id)]]
    delivery_type: BalsamicUmiDeliveryType

    @property
    def enumerated_new_cases(self) -> list[tuple[int, BalsamicUmiCase | ExistingCase]]:
        cases: list[tuple[int, BalsamicUmiCase | ExistingCase]] = []
        for case_index, case in self.enumerated_cases:
            if case.is_new:
                cases.append((case_index, case))
        return cases



"""Functions that deal with modifications of the indexes."""

import logging

from cg.utils.utils import get_hamming_distance

LOG = logging.getLogger(__name__)
DNA_COMPLEMENTS: dict[str, str] = {"A": "T", "C": "G", "G": "C", "T": "A"}
MINIMUM_HAMMING_DISTANCE: int = 3


def is_dual_index(index: str) -> bool:
    """Determines if an index in the raw sample sheet is dual index or not."""
    return "-" in index


def get_reverse_complement_dna_seq(dna: str) -> str:
    """Generates the reverse complement of a DNA sequence."""
    LOG.debug(f"Reverse complement string {dna}")

    return "".join(DNA_COMPLEMENTS[base] for base in reversed(dna))


def get_hamming_distance_index_1(sequence_1: str, sequence_2: str) -> int:
    """
    Get the hamming distance between two index 1 sequences.
    In the case that one sequence is longer than the other, the distance is calculated between
    the shortest sequence and the first segment of equal length of the longest sequence.
    """
    shortest_index_length: int = min(len(sequence_1), len(sequence_2))
    return get_hamming_distance(
        str_1=sequence_1[:shortest_index_length], str_2=sequence_2[:shortest_index_length]
    )


def get_hamming_distance_index_2(
    sequence_1: str, sequence_2: str, is_reverse_complement: bool
) -> int:
    """
    Get the hamming distance between two index 2 sequences.
    In the case that one sequence is longer than the other, the distance is calculated between
    the shortest sequence and the last segment of equal length of the longest sequence.
    If it does not require reverse complement, the calculation is the same as for index 1.
    """
    shortest_index_length: int = min(len(sequence_1), len(sequence_2))
    return (
        get_hamming_distance(
            str_1=sequence_1[-shortest_index_length:], str_2=sequence_2[-shortest_index_length:]
        )
        if is_reverse_complement
        else get_hamming_distance(
            str_1=sequence_1[:shortest_index_length], str_2=sequence_2[:shortest_index_length]
        )
    )



"""Constants for cg."""

from enum import Enum, IntEnum, StrEnum, auto

from cg.utils.date import get_date

VALID_DATA_IN_PRODUCTION = get_date("2017-09-27")

LENGTH_LONG_DATE: int = len("YYYYMMDD")

MAX_ITEMS_TO_RETRIEVE = 50

SCALE_TO_MILLION_READ_PAIRS = 2_000_000
SCALE_TO_READ_PAIRS = 2

ANALYSIS_TYPES = ["tumor_wgs", "tumor_normal_wgs", "tumor_panel", "tumor_normal_panel"]

CAPTUREKIT_CANCER_OPTIONS = (
    "GIcfDNA",
    "GMCKsolid",
    "GMSmyeloid",
    "LymphoMATIC",
    "other (specify in comment field)",
)
CAPTUREKIT_OPTIONS = (
    "Agilent Sureselect CRE",
    "Agilent Sureselect V5",
    "SureSelect Focused Exome",
    "Twist_Target_hg19.bed",
    "other",
)


class JobType(StrEnum):
    UPLOAD = "upload"
    ANALYSIS = "analysis"


class CaseActions(StrEnum):
    ANALYZE = "analyze"
    HOLD = "hold"
    RUNNING = "running"
    TOP_UP = "top-up"

    @classmethod
    def actions(cls) -> list[str]:
        return list(map(lambda action: action.value, cls))


CONTAINER_OPTIONS = ("Tube", "96 well plate", "No container")


class ControlOptions(StrEnum):
    NEGATIVE: str = "negative"
    POSITIVE: str = "positive"
    EMPTY: str = ""


DEFAULT_CAPTURE_KIT = "twistexomecomprehensive_10.2_hg19_design.bed"


class CustomerId(StrEnum):
    CG_INTERNAL_CUSTOMER = "cust000"
    CUST001 = "cust001"
    CUST002 = "cust002"
    CUST003 = "cust003"
    CUST004 = "cust004"
    CUST032 = "cust032"
    CUST042 = "cust042"
    CUST087 = "cust087"
    CUST110 = "cust110"
    CUST127 = "cust127"
    CUST132 = "cust132"
    CUST143 = "cust143"
    CUST147 = "cust147"
    CUST175 = "cust175"
    CUST185 = "cust185"
    CUST201 = "cust201"
    CUST999 = "cust999"


class SequencingRunDataAvailability(StrEnum):
    ON_DISK: str = "ondisk"
    REMOVED: str = "removed"
    REQUESTED: str = "requested"
    PROCESSING: str = "processing"
    RETRIEVED: str = "retrieved"

    @classmethod
    def statuses(cls) -> list[str]:
        return list(map(lambda status: status.value, cls))


class CancerAnalysisType(StrEnum):
    TUMOR_NORMAL = auto()
    TUMOR_NORMAL_PANEL = auto()
    TUMOR_NORMAL_WGS = auto()
    TUMOR_PANEL = auto()
    TUMOR_WGS = auto()


class SexOptions(StrEnum):
    MALE: str = "male"
    FEMALE: str = "female"
    UNKNOWN: str = "unknown"


SARS_COV_REGEX = "^[0-9]{2}CS[0-9]{6}$"

STATUS_OPTIONS = ("affected", "unaffected", "unknown")


class StatusOptions(StrEnum):
    AFFECTED = "affected"
    UNAFFECTED = "unaffected"
    UNKNOWN = "unknown"


class Workflow(StrEnum):
    BALSAMIC = "balsamic"
    BALSAMIC_PON = "balsamic-pon"
    BALSAMIC_UMI = "balsamic-umi"
    DEMULTIPLEX = "demultiplex"
    FLUFFY = "fluffy"
    JASEN = "jasen"
    MICROSALT = "microsalt"
    MIP_DNA = "mip-dna"
    MIP_RNA = "mip-rna"
    MUTANT = "mutant"
    NALLO = "nallo"
    RAREDISEASE = "raredisease"
    RAW_DATA = "raw-data"
    RNAFUSION = "rnafusion"
    RSYNC = "rsync"
    SPRING = "spring"
    TAXPROFILER = "taxprofiler"
    TOMTE = "tomte"


DNA_WORKFLOWS_WITH_SCOUT_UPLOAD: list[Workflow] = [
    Workflow.MIP_DNA,
    Workflow.BALSAMIC,
    Workflow.BALSAMIC_UMI,
]


class FileFormat(StrEnum):
    CSV: str = "csv"
    FASTQ: str = "fastq"
    JSON: str = "json"
    PNG: str = "png"
    TSV: str = "tsv"
    TXT: str = "txt"
    XML: str = "xml"
    YAML: str = "yaml"


class GenomeVersion(StrEnum):
    GRCh37 = "GRCh37"
    GRCh38 = "GRCh38"
    T2T_CHM13 = "T2T-CHM13v2.0"
    CANFAM3 = auto()
    HG19 = "hg19"
    HG38 = "hg38"


class SampleType(StrEnum):
    TUMOR: str = "tumor"
    NORMAL: str = "normal"


class DataDelivery(StrEnum):
    ANALYSIS_FILES = "analysis"
    ANALYSIS_SCOUT = "analysis-scout"
    BAM = "bam"
    FASTQ = "fastq"
    FASTQ_SCOUT = "fastq-scout"
    FASTQ_QC = "fastq_qc"
    FASTQ_ANALYSIS = "fastq-analysis"
    FASTQ_QC_ANALYSIS = "fastq_qc-analysis"
    FASTQ_ANALYSIS_SCOUT = "fastq-analysis-scout"
    NIPT_VIEWER = "nipt-viewer"
    NO_DELIVERY = "no-delivery"
    RAW_DATA_ANALYSIS = "raw_data-analysis"
    RAW_DATA_ANALYSIS_SCOUT = "raw_data-analysis-scout"
    RAW_DATA_SCOUT = "raw_data-scout"
    SCOUT = "scout"
    STATINA = "statina"


class HastaSlurmPartitions(StrEnum):
    DRAGEN: str = "dragen"


class FileExtensions(StrEnum):
    BAM: str = ".bam"
    BCF: str = ".bcf"
    BED: str = ".bed"
    COMPLETE: str = ".complete"
    CONFIG: str = ".config"
    CRAM: str = ".cram"
    CSV: str = ".csv"
    FASTQ: str = ".fastq"
    FASTQ_GZ: str = ".fastq.gz"
    GPG: str = ".gpg"
    GZIP: str = ".gz"
    HTML: str = ".html"
    JSON: str = ".json"
    KEY: str = ".key"
    LOG: str = ".log"
    MD5: str = ".md5"
    MD5SUM: str = ".md5sum"
    NO_EXTENSION: str = ""
    PASS_PHRASE: str = ".passphrase"
    PENDING: str = ".pending"
    PNG: str = ".png"
    SBATCH: str = ".sbatch"
    SPRING: str = ".spring"
    SH: str = ".sh"
    TAR: str = ".tar"
    TMP: str = ".tmp"
    TSV: str = ".tsv"
    TXT: str = ".txt"
    VCF: str = ".vcf"
    VCF_GZ: str = ".vcf.gz"
    XLSX: str = ".xlsx"
    XML: str = ".xml"
    YAML: str = ".yaml"


class APIMethods(StrEnum):
    POST: str = "POST"
    PUT: str = "PUT"
    GET: str = "GET"
    DELETE: str = "DELETE"
    PATCH: str = "PATCH"


class MicrosaltQC:
    AVERAGE_COVERAGE_THRESHOLD: int = 10
    COVERAGE_10X_THRESHOLD: float = 0.75
    DUPLICATION_RATE_THRESHOLD: float = 0.8
    INSERT_SIZE_THRESHOLD: int = 100
    MAPPED_RATE_THRESHOLD: float = 0.3
    NEGATIVE_CONTROL_READS_THRESHOLD: float = 0.2
    TARGET_READS: int = 6000000
    TARGET_READS_FAIL_THRESHOLD: float = 0.7


class MutantQC:
    EXTERNAL_NEGATIVE_CONTROL_READS_THRESHOLD: int = 100000
    INTERNAL_NEGATIVE_CONTROL_READS_THRESHOLD: int = 2000
    FRACTION_OF_SAMPLES_WITH_FAILED_QC_THRESHOLD: float = 0.2
    QUALITY_REPORT_FILE_NAME: str = f"QC_report{FileExtensions.JSON}"


DRY_RUN_MESSAGE = "Dry run: process call will not be executed!"


class MetaApis:
    ANALYSIS_API: str = "analysis_api"


class WorkflowManager(StrEnum):
    Slurm = "slurm"
    Tower = "nf_tower"


class Strandedness(StrEnum):
    """Strandedness types."""

    FORWARD = "forward"
    REVERSE = "reverse"
    UNSTRANDED = "unstranded"


class ReadDirection(IntEnum):
    """Read direction types."""

    FORWARD: int = 1
    REVERSE: int = 2


PIPELINES_USING_PARTIAL_ANALYSES: list[Workflow] = [Workflow.MICROSALT, Workflow.MUTANT]


class MultiQC(StrEnum):
    """MultiQC constants"""

    MULTIQC: str = "multiqc"
    MULTIQC_DATA: str = "multiqc_data"


NG_UL_SUFFIX: str = " ng/uL"


class SequencingQCStatus(Enum):
    FAILED = auto()
    PASSED = auto()
    PENDING = auto()


class SampleStatus(StrEnum):
    INCOMING = "incoming"
    LABPREP = "labprep"
    SEQUENCING = "sequencing"

    @classmethod
    def statuses(cls) -> list[str]:
        return [cls.INCOMING, cls.LABPREP, cls.SEQUENCING]



import logging
from datetime import datetime

from cg.constants import DataDelivery, Sex
from cg.services.orders.constants import ORDER_TYPE_WORKFLOW_MAP
from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.storing.service import StoreOrderService
from cg.services.orders.validation.order_types.microsalt.models.order import MicrosaltOrder
from cg.services.orders.validation.order_types.microsalt.models.sample import MicrosaltSample
from cg.services.orders.validation.order_types.mutant.models.order import MutantOrder
from cg.services.orders.validation.order_types.mutant.models.sample import MutantSample
from cg.store.models import ApplicationVersion
from cg.store.models import Case as DbCase
from cg.store.models import CaseSample, Customer
from cg.store.models import Order as DbOrder
from cg.store.models import Organism
from cg.store.models import Sample as DbSample
from cg.store.store import Store

LOG = logging.getLogger(__name__)

MicrobialOrder = MicrosaltOrder | MutantOrder
MicrobialSample = MicrosaltSample | MutantSample


class StoreMicrobialOrderService(StoreOrderService):
    """
    Storing service for microbial orders.
    These include:
    - Mutant samples
    - Microsalt samples
    """

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status = status_db
        self.lims = lims_service

    def store_order(self, order: MicrobialOrder) -> dict:
        self._fill_in_sample_verified_organism(order.samples)
        project_data, lims_map = self.lims.process_lims(
            samples=order.samples,
            customer=order.customer,
            ticket=order._generated_ticket_id,
            order_name=order.name,
            workflow=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
            delivery_type=DataDelivery(order.delivery_type),
            skip_reception_control=order.skip_reception_control,
        )
        self._fill_in_sample_ids(samples=order.samples, lims_map=lims_map)

        samples = self.store_order_data_in_status_db(order)
        return {"project": project_data, "records": samples}

    def store_order_data_in_status_db(self, order: MicrobialOrder) -> list[DbSample]:
        """Store microbial samples in the status database."""

        customer: Customer = self.status.get_customer_by_internal_id(order.customer)
        new_samples = []
        db_order: DbOrder = self.status.add_order(
            customer=customer,
            ticket_id=order._generated_ticket_id,
        )
        db_case: DbCase = self._create_case(customer=customer, order=order)

        with self.status.no_autoflush_context():
            for sample in order.samples:
                organism: Organism = self._ensure_organism(sample)
                db_sample = self._create_db_sample(
                    customer=customer,
                    order_name=order.name,
                    organism=organism,
                    sample=sample,
                    ticket_id=order._generated_ticket_id,
                )
                link: CaseSample = self.status.relate_sample(
                    case=db_case, sample=db_sample, status="unknown"
                )
                self.status.add_item_to_store(link)
                new_samples.append(db_sample)
            db_order.cases.append(db_case)

        self.status.add_item_to_store(db_case)
        self.status.add_item_to_store(db_order)
        self.status.add_multiple_items_to_store(new_samples)
        self.status.commit_to_store()
        return new_samples

    def _fill_in_sample_verified_organism(self, samples: list[MicrobialSample]):
        for sample in samples:
            organism_id = sample.organism
            reference_genome = sample.reference_genome
            organism: Organism = self.status.get_organism_by_internal_id(internal_id=organism_id)
            is_verified: bool = (
                organism and organism.reference_genome == reference_genome and organism.verified
            )
            sample._verified_organism = is_verified

    def _create_case(self, customer: Customer, order: MicrobialOrder) -> DbCase:
        case: DbCase = self.status.add_case(
            data_analysis=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
            data_delivery=DataDelivery(order.delivery_type),
            name=str(order._generated_ticket_id),
            panels=None,
            ticket=str(order._generated_ticket_id),
            priority=order.samples[0].priority,
        )
        case.customer = customer
        return case

    def _ensure_organism(self, sample: MicrobialSample) -> Organism:
        organism: Organism = self.status.get_organism_by_internal_id(sample.organism)
        if not organism:
            organism: Organism = self.status.add_organism(
                internal_id=sample.organism,
                name=sample.organism,
                reference_genome=sample.reference_genome,
            )
            self.status.add_item_to_store(organism)
            self.status.commit_to_store()
        return organism

    def _create_db_sample(
        self,
        customer: Customer,
        order_name: str,
        organism: Organism,
        sample: MicrobialSample,
        ticket_id: int,
    ) -> DbSample:
        application_tag: str = sample.application
        application_version: ApplicationVersion = (
            self.status.get_current_application_version_by_tag(tag=application_tag)
        )
        return self.status.add_sample(
            name=sample.name,
            sex=Sex.UNKNOWN,
            comment=sample.comment,
            control=sample.control,
            internal_id=sample._generated_lims_id,
            order=order_name,
            ordered=datetime.now(),
            original_ticket=str(ticket_id),
            priority=sample.priority,
            application_version=application_version,
            customer=customer,
            organism=organism,
            reference_genome=sample.reference_genome,
        )



import logging
from datetime import date
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

from cg.clients.freshdesk.freshdesk_client import FreshdeskClient
from cg.clients.freshdesk.models import TicketCreate, TicketResponse
from cg.constants.priority import Priority
from cg.meta.orders.utils import get_due_by_date, get_ticket_status, get_ticket_tags
from cg.models.orders.constants import OrderType
from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.store.models import Case, Customer, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class TicketHandler:
    """Handle tickets in the meta orders context"""

    NEW_LINE = "<br />"

    def __init__(self, db: Store, client: FreshdeskClient, system_email_id: int, env: str):
        self.client: FreshdeskClient = client
        self.status_db: Store = db
        self.system_email_id: int = system_email_id
        self.env: str = env

    def create_ticket(
        self, order: Order, user_name: str, user_mail: str, order_type: OrderType
    ) -> int | None:
        """Create a ticket and return the ticket number"""
        message: str = self.create_new_ticket_header(
            message=self.create_xml_sample_list(order=order, user_name=user_name),
            order=order,
            order_type=order_type,
        )

        tags: list[str] = get_ticket_tags(
            order=order, order_type=order_type, status_db=self.status_db
        )
        status: int = get_ticket_status(order=order)

        order_priority: Priority = self._get_max_case_priority(order=order)
        deadline_date: date = get_due_by_date(priority=order_priority)

        with TemporaryDirectory() as temp_dir:
            attachments: Path = self.create_attachment_file(order=order, temp_dir=temp_dir)

            freshdesk_ticket = TicketCreate(
                email=user_mail,
                description=message,
                email_config_id=self.system_email_id,
                name=user_name,
                subject=order.name,
                type="Order",
                tags=tags,
                custom_fields={
                    "cf_environment": self.env,
                },
                attachments=[],
                status=status,
                due_by=deadline_date,
                fr_due_by=deadline_date,
            )
            ticket_response: TicketResponse = self.client.create_ticket(
                ticket=freshdesk_ticket, attachments=[attachments]
            )
            LOG.info(f"{ticket_response.id}: opened new ticket in Freshdesk")

        return ticket_response.id

    def create_attachment_file(self, order: Order, temp_dir: str) -> Path:
        """Create a single attachment file for the ticket"""
        order_file_path = Path(temp_dir) / "order.json"
        with order_file_path.open("w") as order_file:
            order_file.write(order.json())
        return order_file_path

    def create_xml_sample_list(self, order: Order, user_name: str) -> str:
        message = ""
        if isinstance(order, OrderWithCases):
            message = self.create_case_xml_sample_list(order=order, message=message)
        else:
            for sample in order.samples:
                message = self.add_sample_name_to_message(message=message, sample_name=sample.name)
                message = self.add_sample_apptag_to_message(
                    message=message, application=sample.application
                )
                message = self.add_sample_priority_to_message(
                    message=message, priority=sample.priority
                )
                message = self.add_sample_comment_to_message(
                    message=message, comment=sample.comment
                )

        message += self.NEW_LINE
        message = self.add_order_comment_to_message(message=message, comment=order.comment)
        message = self.add_user_name_to_message(message=message, name=user_name)
        message = self.add_customer_to_message(message=message, customer_id=order.customer)

        return message

    @staticmethod
    def create_new_ticket_header(message: str, order: Order, order_type: OrderType) -> str:
        nr_samples = (
            len(order.samples)
            if isinstance(order, OrderWithSamples)
            else len(order.enumerated_new_samples)
        )
        return f"New order with {nr_samples} new {order_type} samples:" + message

    def add_sample_name_to_message(self, message: str, sample_name: str) -> str:
        message += f"{self.NEW_LINE}{sample_name}"
        return message

    @staticmethod
    def add_sample_apptag_to_message(message: str, application: str | None) -> str:
        if application:
            message += f", application: {application}"
        return message

    @staticmethod
    def add_sample_case_name_to_message(message: str, case_name: str | None) -> str:
        if case_name:
            message += f", case: {case_name}"
        return message

    def add_existing_sample_info_to_message(
        self, message: str, customer_id: str, internal_id: str, case_name: str
    ) -> str:
        existing_sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=internal_id)

        sample_customer = ""
        if existing_sample.customer_id != customer_id:
            sample_customer = " from " + existing_sample.customer.internal_id
        message += f"{existing_sample.name}, application: {existing_sample.application_version.application.tag}, case: {case_name} (already existing sample{sample_customer}), priority: {existing_sample.priority.name.lower()}"
        return message

    @staticmethod
    def add_sample_priority_to_message(message: str, priority: str | None) -> str:
        if priority:
            message += f", priority: {priority}"
        return message

    @staticmethod
    def add_sample_comment_to_message(message: str, comment: str | None) -> str:
        if comment:
            message += f", {comment}"
        return message

    def add_order_comment_to_message(self, message: str, comment: str | None) -> str:
        if comment:
            message += f"{self.NEW_LINE}{comment}."
        return message

    def add_user_name_to_message(self, message: str, name: str | None) -> str:
        if name:
            message += f"{self.NEW_LINE}{name}"
        return message

    def add_customer_to_message(self, message: str, customer_id: str) -> str:
        """Add customer info to message and return updated message."""
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        message += f", {customer.name} ({customer_id})"
        return message

    @classmethod
    def replace_empty_string_with_none(cls, obj: Any) -> Any:
        """Recursive function that replaces empty string in nested dicts/lists with None"""
        if obj == "":
            return None
        if isinstance(obj, dict):
            for key, item in obj.items():
                if isinstance(item, list):
                    processed_list = [
                        cls.replace_empty_string_with_none(list_item) for list_item in item
                    ]
                    obj[key] = processed_list
                else:
                    obj[key] = cls.replace_empty_string_with_none(item)
        return obj

    def create_case_xml_sample_list(self, order, message: str) -> str:
        for case in order.cases:
            if not case.is_new:
                db_case = self.status_db.get_case_by_internal_id(case.internal_id)
                for sample in db_case.samples:
                    message += self.NEW_LINE
                    message = self.add_existing_sample_info_to_message(
                        message=message,
                        customer_id=sample.customer.internal_id,
                        internal_id=sample.internal_id,
                        case_name=db_case.name,
                    )
            else:
                for sample in case.samples:
                    if not sample.is_new:
                        message += self.NEW_LINE
                        message = self.add_existing_sample_info_to_message(
                            message=message,
                            customer_id=order.customer,
                            internal_id=sample.internal_id,
                            case_name=case.name,
                        )
                    else:
                        message = self.add_sample_name_to_message(
                            message=message, sample_name=sample.name
                        )
                        message = self.add_sample_apptag_to_message(
                            message=message, application=sample.application
                        )
                        message = self.add_sample_case_name_to_message(
                            message=message, case_name=case.name
                        )
                        message = self.add_sample_priority_to_message(
                            message=message, priority=case.priority
                        )
                        message = self.add_sample_comment_to_message(
                            message=message, comment=sample.comment
                        )
        return message

    def _get_max_case_priority(self, order: Order) -> Priority:
        """Get max case priority for a given order."""
        priority_list: list[Priority] = []

        if isinstance(order, OrderWithCases):
            for index, new_case in order.enumerated_new_cases:
                priority_list.append(Priority[new_case.priority])

            for index, case in order.enumerated_existing_cases:
                case: Case = self.status_db.get_case_by_internal_id(case.internal_id)
                priority_list.append(case.priority)

        if isinstance(order, OrderWithSamples):
            for sample in order.samples:
                priority_list.append(Priority[sample.priority])

        return max(priority_list)



from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import Workflow
from cg.meta.workflow.fastq import BalsamicFastqHandler, MicrosaltFastqHandler, MipFastqHandler
from cg.models.cg_config import CGConfig, CommonAppConfig
from cg.services.analysis_starter.configurator.configurator import Configurator
from cg.services.analysis_starter.configurator.extensions.nallo import NalloExtension
from cg.services.analysis_starter.configurator.extensions.pipeline_extension import (
    PipelineExtension,
)
from cg.services.analysis_starter.configurator.extensions.raredisease import RarediseaseExtension
from cg.services.analysis_starter.configurator.extensions.tomte_extension import TomteExtension
from cg.services.analysis_starter.configurator.file_creators.balsamic_config import (
    BalsamicConfigFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.gene_panel import GenePanelFileCreator
from cg.services.analysis_starter.configurator.file_creators.managed_variants import (
    ManagedVariantsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.microsalt_config import (
    MicrosaltConfigFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.mip_dna_config import (
    MIPDNAConfigFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.config_file import (
    NextflowConfigFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.abstract import (
    ParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.nallo import (
    NalloParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.raredisease import (
    RarediseaseParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.rnafusion import (
    RNAFusionParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.taxprofiler import (
    TaxprofilerParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.tomte_params_file_creator import (
    TomteParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.nallo_sample_sheet_creator import (
    NalloSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.protocol import (
    SampleSheetCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.raredisease_sample_sheet_creator import (
    RarediseaseSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.rnafusion_sample_sheet_creator import (
    RNAFusionSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.taxprofiler_sample_sheet_creator import (
    TaxprofilerSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.tomte_sample_sheet_creator import (
    TomteSampleSheetCreator,
)
from cg.services.analysis_starter.configurator.implementations.balsamic import BalsamicConfigurator
from cg.services.analysis_starter.configurator.implementations.microsalt import (
    MicrosaltConfigurator,
)
from cg.services.analysis_starter.configurator.implementations.mip_dna import MIPDNAConfigurator
from cg.services.analysis_starter.configurator.implementations.nextflow import NextflowConfigurator
from cg.store.store import Store


class ConfiguratorFactory:

    def __init__(self, cg_config: CGConfig):
        self.cg_config = cg_config
        self.housekeeper_api: HousekeeperAPI = cg_config.housekeeper_api
        self.lims_api: LimsAPI = cg_config.lims_api
        self.store: Store = cg_config.status_db

    def get_configurator(self, workflow: Workflow) -> Configurator:
        match workflow:
            case Workflow.BALSAMIC | Workflow.BALSAMIC_UMI:
                return self._get_balsamic_configurator()
            case Workflow.MICROSALT:
                return self._get_microsalt_configurator()
            case Workflow.MIP_DNA:
                return self._get_mip_dna_configurator()
            case (
                Workflow.NALLO
                | Workflow.RAREDISEASE
                | Workflow.RNAFUSION
                | Workflow.TAXPROFILER
                | Workflow.TOMTE
            ):
                return self._get_nextflow_configurator(workflow)
            case _:
                raise NotImplementedError

    def _get_nextflow_configurator(self, workflow: Workflow) -> NextflowConfigurator:
        config_file_creator: NextflowConfigFileCreator = self._get_nextflow_config_file_creator(
            workflow
        )
        params_file_creator: ParamsFileCreator = self._get_params_file_creator(workflow)
        sample_sheet_creator: SampleSheetCreator = self._get_sample_sheet_creator(workflow)
        extension: PipelineExtension = self._get_pipeline_extension(workflow)
        return NextflowConfigurator(
            config_file_creator=config_file_creator,
            params_file_creator=params_file_creator,
            pipeline_config=self._get_pipeline_config(workflow),
            sample_sheet_creator=sample_sheet_creator,
            store=self.store,
            pipeline_extension=extension,
        )

    def _get_nextflow_config_file_creator(self, workflow: Workflow) -> NextflowConfigFileCreator:
        pipeline_config: CommonAppConfig = self._get_pipeline_config(workflow)
        return NextflowConfigFileCreator(
            account=pipeline_config.slurm.account,
            platform=pipeline_config.platform,
            resources=pipeline_config.resources,
            store=self.store,
            workflow_config_path=pipeline_config.config,
        )

    def _get_params_file_creator(self, workflow: Workflow) -> ParamsFileCreator:
        params: str = self._get_pipeline_config(workflow).params
        match workflow:
            case Workflow.NALLO:
                return NalloParamsFileCreator(params)
            case Workflow.RAREDISEASE:
                return RarediseaseParamsFileCreator(
                    lims=self.lims_api, params=params, store=self.store
                )
            case Workflow.RNAFUSION:
                return RNAFusionParamsFileCreator(params)
            case Workflow.TAXPROFILER:
                return TaxprofilerParamsFileCreator(params)
            case Workflow.TOMTE:
                return TomteParamsFileCreator(
                    lims_api=self.lims_api, params=params, status_db=self.store
                )
            case _:
                raise NotImplementedError(f"There is no params file creator for {workflow}")

    def _get_pipeline_config(self, workflow: Workflow) -> CommonAppConfig:
        return getattr(self.cg_config, workflow)

    def _get_sample_sheet_creator(self, workflow: Workflow) -> SampleSheetCreator:
        match workflow:
            case Workflow.NALLO:
                return NalloSampleSheetCreator(
                    housekeeper_api=self.housekeeper_api, status_db=self.store
                )
            case Workflow.RAREDISEASE:
                return RarediseaseSampleSheetCreator(
                    housekeeper_api=self.cg_config.housekeeper_api,
                    store=self.store,
                )
            case Workflow.RNAFUSION:
                return RNAFusionSampleSheetCreator(
                    housekeeper_api=self.housekeeper_api, store=self.store
                )
            case Workflow.TAXPROFILER:
                return TaxprofilerSampleSheetCreator(
                    housekeeper_api=self.housekeeper_api, store=self.store
                )
            case Workflow.TOMTE:
                return TomteSampleSheetCreator(
                    housekeeper_api=self.housekeeper_api, store=self.store
                )
            case _:
                raise NotImplementedError(f"No sample sheet creator implemented for {workflow}")

    def _get_pipeline_extension(self, workflow: Workflow) -> PipelineExtension:
        match workflow:
            case Workflow.NALLO:
                gene_panel_creator: GenePanelFileCreator = self._get_gene_panel_file_creator(
                    workflow
                )
                return NalloExtension(gene_panel_file_creator=gene_panel_creator)
            case Workflow.RAREDISEASE:
                gene_panel_creator: GenePanelFileCreator = self._get_gene_panel_file_creator(
                    workflow
                )
                managed_variants_creator: ManagedVariantsFileCreator = (
                    self._get_managed_variants_file_creator(workflow)
                )
                return RarediseaseExtension(
                    gene_panel_file_creator=gene_panel_creator,
                    managed_variants_file_creator=managed_variants_creator,
                )
            case Workflow.TOMTE:
                gene_panel_creator: GenePanelFileCreator = self._get_gene_panel_file_creator(
                    workflow
                )
                return TomteExtension(gene_panel_file_creator=gene_panel_creator)
            case _:
                return PipelineExtension()

    def _get_gene_panel_file_creator(self, workflow: Workflow) -> GenePanelFileCreator:
        return GenePanelFileCreator(scout_api=self._get_scout_api(workflow), store=self.store)

    def _get_managed_variants_file_creator(self, workflow: Workflow) -> ManagedVariantsFileCreator:
        return ManagedVariantsFileCreator(scout_api=self._get_scout_api(workflow), store=self.store)

    def _get_scout_api(self, workflow: Workflow) -> ScoutAPI:
        return (
            self.cg_config.scout_api_38
            if workflow == Workflow.NALLO
            else self.cg_config.scout_api_37
        )

    def _get_balsamic_configurator(self) -> BalsamicConfigurator:
        return BalsamicConfigurator(
            config_file_creator=BalsamicConfigFileCreator(
                cg_balsamic_config=self.cg_config.balsamic,
                lims_api=self.lims_api,
                status_db=self.store,
            ),
            fastq_handler=BalsamicFastqHandler(
                housekeeper_api=self.housekeeper_api,
                root_dir=Path(self.cg_config.balsamic.root),
                status_db=self.store,
            ),
            config=self.cg_config.balsamic,
            store=self.store,
        )

    def _get_microsalt_configurator(self) -> MicrosaltConfigurator:
        return MicrosaltConfigurator(
            config_file_creator=self._get_microsalt_config_file_creator(),
            fastq_handler=MicrosaltFastqHandler(
                housekeeper_api=self.housekeeper_api,
                root_dir=Path(self.cg_config.microsalt.root),
                status_db=self.store,
            ),
            lims_api=self.lims_api,
            microsalt_config=self.cg_config.microsalt,
            store=self.store,
        )

    def _get_microsalt_config_file_creator(self) -> MicrosaltConfigFileCreator:
        return MicrosaltConfigFileCreator(
            lims_api=self.lims_api,
            queries_path=self.cg_config.microsalt.queries_path,
            store=self.store,
        )

    def _get_mip_dna_configurator(self) -> MIPDNAConfigurator:
        root: str = self.cg_config.mip_rd_dna.root
        return MIPDNAConfigurator(
            cg_mip_config=self.cg_config.mip_rd_dna,
            config_file_creator=self._get_mip_dna_config_file_creator(root=root),
            fastq_handler=MipFastqHandler(
                self.housekeeper_api, root_dir=Path(root), status_db=self.store
            ),
            gene_panel_file_creator=self._get_gene_panel_file_creator(Workflow.MIP_DNA),
            managed_variants_file_creator=self._get_managed_variants_file_creator(Workflow.MIP_DNA),
            store=self.store,
        )

    def _get_mip_dna_config_file_creator(self, root: str) -> MIPDNAConfigFileCreator:
        return MIPDNAConfigFileCreator(lims_api=self.lims_api, root=root, store=self.store)
